@@ -1,4 +1,4 @@
-### Copyright (C) 2002-2003 Stephen Kennedy <stevek@gnome.org>
+### Copyright (C) 2002-2004 Stephen Kennedy <stevek@gnome.org>
 
 ### This program is free software; you can redistribute it and/or modify
 ### it under the terms of the GNU General Public License as published by
@@ -215,7 +215,6 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         gnomeglade.Component.__init__(self, paths.share_dir("glade2/dirdiff.glade"), "dirdiff")
         self.toolbar.set_style( self.prefs.get_toolbar_style() )
         self._map_widgets_into_lists( ["treeview", "fileentry", "diffmap", "scrolledwindow", "linkmap"] )
-        self.lock = 0
         self.popup_menu = DirDiffMenu(self)
         self.set_num_panes(num_panes)
         self.on_treeview_focus_out_event(None, None)
@@ -248,7 +247,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         for r in [ misc.ListItem(i) for i in self.prefs.regexes.split("\n") ]:
             if r.active:
                 try:
-                    self.regexes.append( re.compile(r.value) )
+                    self.regexes.append( re.compile(r.value+"(?m)") )
                 except re.error, e:
                     misc.run_dialog(
                         text=_("Error converting pattern '%s' to regular expression") % r.value )
@@ -271,13 +270,18 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                 func = lambda x, r=cregex : r.match(x) == None
                 self.name_filters_available.append( TypeFilter(f.name, f.active, func) )
         self.name_filters = []
+        tips = gtk.Tooltips()
         for i,f in misc.enumerate(self.name_filters_available):
             icon = gtk.Image()
             icon.set_from_stock(gtk.STOCK_FIND, gtk.ICON_SIZE_LARGE_TOOLBAR)
-            icon.show()
-            toggle = self.toolbar.append_element(gtk.TOOLBAR_CHILD_TOGGLEBUTTON, None, f.label,
-                _("Hide %s") % f.label, "", icon, self._update_name_filter, i )
+            toggle = gtk.ToggleToolButton()
+            toggle.set_property("label",f.label)
+            toggle.set_icon_widget(icon)
+            toggle.connect("toggled", lambda b,i=i : self._update_name_filter(b,i) )
             toggle.set_active(f.active)
+            self.toolbar.insert(toggle, -1)
+            toggle.show_all()
+            toggle.set_tooltip(tips, _("Hide %s") % f.label )
 
     def on_preference_changed(self, key, value):
         if key == "toolbar_style":
@@ -286,14 +290,14 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             self.update_regexes()
 
     def _do_to_others(self, master, objects, methodname, args):
-        if self.lock == 0:
-            self.lock = 1
+        if not hasattr(self, "do_to_others_lock"):
+            self.do_to_others_lock = 1
             try:
                 for o in filter(lambda x:x!=master, objects[:self.num_panes]):
                     method = getattr(o,methodname)
                     method(*args)
             finally:
-                self.lock = 0
+                delattr(self, "do_to_others_lock")
 
     def _sync_vscroll(self, adjustment):
         adjs = map(lambda x: x.get_vadjustment(), self.scrolledwindow)
@@ -319,14 +323,14 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             self._update_item_state(iter)
         else: # nope its gone
             self.model.remove(iter)
-        self._update_difmaps()
+        self._update_diffmaps()
 
     def file_created(self, path, pane):
         iter = self.model.get_iter(path)
         while iter and self.model.get_path(iter) != (0,):
             self._update_item_state( iter )
             iter = self.model.iter_parent(iter)
-        self._update_difmaps()
+        self._update_diffmaps()
 
     def on_fileentry_activate(self, entry):
         locs = [ e.get_full_path(0) for e in self.fileentry[:self.num_panes] ]
@@ -356,6 +360,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         self.scheduler.add_task( self._search_recursively_iter( path ).next )
 
     def _search_recursively_iter(self, rootpath):
+        self.filter_hide_current.set_sensitive(False)
         yield _("[%s] Scanning") % self.label_text
         prefixlen = 1 + len( self.model.value_path( self.model.get_iter(rootpath), 0 ) )
         symlinks_followed = {} # only follow symlinks once
@@ -404,7 +409,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                                 self.items[ ci ][pane] = i
                     def get(self):
                         if len(self.bad):
-                            misc.run_dialog(_("You are running a case insensitve comparison on"
+                            misc.run_dialog(_("You are running a case insensitive comparison on"
                                 " a case sensitive filesystem. Some files are not visible:\n%s")
                                 % "\n".join( self.bad ), self.parent )
                         keys = self.items.keys()
@@ -471,6 +476,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                     level += 1
                     self.treeview[0].expand_row( path[:level], 0)
         yield _("[%s] Done") % self.label_text
+        self.filter_hide_current.set_sensitive(True)
 
     def launch_comparison(self, iter, pane, force=1):
         """Launch comparison at 'iter'. 
@@ -623,11 +629,11 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
 
     def on_treeview_row_expanded(self, view, iter, path):
         self._do_to_others(view, self.treeview, "expand_row", (path,0) )
-        self._update_difmaps()
+        self._update_diffmaps()
 
     def on_treeview_row_collapsed(self, view, me, path):
         self._do_to_others(view, self.treeview, "collapse_row", (path,) )
-        self._update_difmaps()
+        self._update_diffmaps()
 
     def on_treeview_focus_in_event(self, tree, event):
         self.treeview_focussed = tree
@@ -781,9 +787,6 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                 self.model.set_state(iter, j,  tree.STATE_MISSING)
         return different
 
-    def update_diff_maps(self):
-        return
-
     def on_treeview_button_press_event(self, treeview, event):
         # unselect other panes
         for t in filter(lambda x:x!=treeview, self.treeview[:self.num_panes]):
@@ -831,7 +834,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         self.label_text = " : ".join(shortnames)
         self.label_changed()
 
-    def _update_difmaps(self):
+    def _update_diffmaps(self):
         self.diffmap[0].queue_draw()
         self.diffmap[1].queue_draw()
 
@@ -969,4 +972,24 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         for path in changed_paths:
             self._update_item_state( model.get_iter(path) )
 
+    def next_diff(self, direction):
+        if self.treeview_focussed:
+            pane = self.treeview.index( self.treeview_focussed )
+        else:
+            pane = 0
+        start_iter = self.model.get_iter( (self._get_selected_paths(pane) or [(0,)])[-1] )
+
+        def goto_iter(it):
+            curpath = self.model.get_path(it)
+            for i in range(len(curpath)-1):
+                self.treeview[pane].expand_row( curpath[:i+1], 0)
+            self.treeview[pane].set_cursor(curpath)
+
+        search = {gdk.SCROLL_UP : self.model.inorder_search_up}.get(direction, self.model.inorder_search_down)
+        for it in search( start_iter ):
+            state = int(self.model.get_state( it, pane ))
+            if state not in (tree.STATE_NORMAL, tree.STATE_EMPTY):
+                goto_iter(it)
+                return
+
 gobject.type_register(DirDiff)
