@@ -1,4 +1,4 @@
-### Copyright (C) 2002-2003 Stephen Kennedy <steve9000@users.sf.net>
+### Copyright (C) 2002-2004 Stephen Kennedy <stevek@gnome.org>
 
 ### This program is free software; you can redistribute it and/or modify
 ### it under the terms of the GNU General Public License as published by
@@ -30,47 +30,71 @@ import misc
 import melddoc
 import filediff
 import cvsview
+import svnview
 import dirdiff
 import task
 
-version = "0.9.2"
+# optional
+sourceview_available = 0
+
+for sourceview in "gtksourceview sourceview".split():
+    try:
+        __import__(sourceview)
+        sourceview_available = 1
+        break
+    except ImportError:
+        pass
+
+version = "0.9.5"
 
 # magic developer switch, changes some behaviour
 developer = 0
 
 ################################################################################
 #
-# BrowseFileDialog
+# NewDocDialog
 #
 ################################################################################
 
-class BrowseFileDialog(gnomeglade.Component):
-    def __init__(self, parentapp, labels, callback, isdir=0):
-        gnomeglade.Component.__init__(self, paths.share_dir("glade2/meldapp.glade"), "browsefile")
+class NewDocDialog(gnomeglade.Component):
+
+    TYPE = misc.struct(DIFF2=0, DIFF3=1, DIR2=2, DIR3=3, CVS=4, SVN=6)
+         
+    def __init__(self, parentapp, type):
+        self.parentapp = parentapp
+        gnomeglade.Component.__init__(self, paths.share_dir("glade2/meldapp.glade"), "newdialog")
+        self._map_widgets_into_lists( ("fileentry", "direntry", "cvsentry", "svnentry", "three_way_compare", "tablabel") )
+        self.entrylists = self.fileentry, self.direntry, self.cvsentry, self.svnentry
         self.widget.set_transient_for(parentapp.widget)
-        self.numfile = len(labels)
-        self.callback = callback
-        self.labels = map(gtk.Label, labels )
-        self.entries = [ gnomeglade.FileEntry("fileentry", _("Browse %s") % x) for x in labels ]
-        for i in range(self.numfile):
-            self.labels[i].set_justify(gtk.JUSTIFY_RIGHT)
-            self.table.attach(self.labels[i], 0, 1, i, i+1, gtk.SHRINK)
-            self.entries[i].set_directory_entry(isdir)
-            self.entries[i].connect("activate", self.on_activate)
-            self.table.attach(self.entries[i] , 1, 2, i, i+1)
+        cur_page = type // 2
+        self.notebook.set_current_page( cur_page )
         self.widget.show_all()
 
-    def on_activate(self, entry):
-        i = self.entries.index(entry)
-        if i == self.numfile - 1:
-            self.button_ok.activate()
-        else:
-            self.entries[i+1].gtk_entry().grab_focus()
+    def on_entry_activate(self, entry):
+        for el in self.entrylists:
+            if entry in el:
+                i = el.index(entry)
+                if i == len(el) - 1:
+                    self.button_ok.grab_focus()
+                else:
+                    el[i+1].gtk_entry().grab_focus()
+
+    def on_three_way_toggled(self, button):
+        page = self.three_way_compare.index(button)
+        self.entrylists[page][0].set_sensitive( button.get_active() )
+        self.entrylists[page][ not button.get_active() ].gtk_entry().grab_focus()
 
     def on_response(self, dialog, arg):
         if arg==gtk.RESPONSE_OK:
-            files = [e.get_full_path(0) or "" for e in self.entries]
-            self.callback(files)
+            page = self.notebook.get_current_page()
+            paths = [ e.get_full_path(0) or "" for e in self.entrylists[page] ]
+            if page < 2 and not self.three_way_compare[page].get_active():
+                paths.pop(0)
+            methods = (self.parentapp.append_filediff,
+                       self.parentapp.append_dirdiff,
+                       self.parentapp.append_cvsview,
+                       self.parentapp.append_svnview )
+            methods[page](paths)
         self.widget.destroy()
 
 ################################################################################
@@ -209,7 +233,10 @@ class PreferencesDialog(gnomeglade.Component):
             self.radiobutton_gnome_font.set_active(1)
         self.fontpicker.set_font_name( self.prefs.custom_font )
         self.spinbutton_tabsize.set_value( self.prefs.tab_size )
+        self.option_wrap_lines.set_history( self.prefs.edit_wrap_lines )
         self.checkbutton_supply_newline.set_active( self.prefs.supply_newline )
+        self.checkbutton_show_line_numbers.set_active( self.prefs.show_line_numbers )
+        self.checkbutton_use_syntax_highlighting.set_active( self.prefs.use_syntax_highlighting )
         self.editor_command[ self.editor_radio_values.get(self.prefs.edit_command_type, "internal") ].set_active(1)
         self.gnome_default_editor_label.set_text( "(%s)" % " ".join(self.prefs.get_gnome_editor_command([])) )
         self.custom_edit_command_entry.set_text( " ".join(self.prefs.get_custom_editor_command([])) )
@@ -226,6 +253,7 @@ class PreferencesDialog(gnomeglade.Component):
         cols = [ ("Name", type("")), ("Active", type(0)), ("Regex", type("")) ]
         self.textfilter = ListWidget( cols, self.prefs, "regexes")
         self.text_filters_box.pack_start(self.textfilter.widget)
+        self.checkbutton_ignore_blank_lines.set_active( self.prefs.ignore_blank_lines )
         # encoding
         self.entry_text_codecs.set_text( self.prefs.text_codecs )
         self._map_widgets_into_lists( ["save_encoding"] )
@@ -256,8 +284,18 @@ class PreferencesDialog(gnomeglade.Component):
             self.prefs.use_custom_font = custom
     def on_spinbutton_tabsize_changed(self, spin):
         self.prefs.tab_size = int(spin.get_value())
+    def on_option_wrap_lines_changed(self, option):
+        self.prefs.edit_wrap_lines = option.get_history()
     def on_checkbutton_supply_newline_toggled(self, check):
         self.prefs.supply_newline = check.get_active()
+    def on_checkbutton_show_line_numbers_toggled(self, check):
+        self.prefs.show_line_numbers = check.get_active()
+        if check.get_active() and not sourceview_available:
+            misc.run_dialog(_("Line numbers are only available if you have pysourceview installed.") )
+    def on_checkbutton_use_syntax_highlighting_toggled(self, check):
+        self.prefs.use_syntax_highlighting = check.get_active()
+        if check.get_active() and not sourceview_available:
+            misc.run_dialog(_("Syntax highlighting is only available if you have pysourceview installed.") )
     def on_editor_command_toggled(self, radio):
         if radio.get_active():
             idx = self.editor_command.index(radio)
@@ -277,6 +315,8 @@ class PreferencesDialog(gnomeglade.Component):
     #
     # filters
     #
+    def on_checkbutton_ignore_blank_lines_toggled(self, check):
+        self.prefs.ignore_blank_lines = check.get_active()
 
     #
     # encoding
@@ -292,7 +332,7 @@ class PreferencesDialog(gnomeglade.Component):
     def on_cvs_compression_toggled(self, toggle):
         self.prefs.cvs_compression = toggle.get_active()
     def on_cvs_compression_value_changed(self, spin):
-        self.prefs.cvs_compression_value = spin.get_value()
+        self.prefs.cvs_compression_value = int(spin.get_value())
     def on_cvs_ignore_cvsrc_toggled(self, toggle):
         self.prefs.cvs_ignore_cvsrc = toggle.get_active()
     def on_cvs_binary_activate(self, fileentry):
@@ -315,48 +355,18 @@ class PreferencesDialog(gnomeglade.Component):
 
 class MeldStatusBar(object):
 
-    def __init__(self, appbar):
-        self.appbar = appbar
-        self.statusmessages = []
-        self.statuscount = []
-        self.appbar.set_default("OK")
+    def __init__(self, task_progress, task_status, doc_status):
+        self.task_progress = task_progress
+        self.task_status = task_status
+        self.doc_status = doc_status
 
-    def add_status(self, status, timeout=4000, allow_duplicate=0):
-        if not allow_duplicate:
-            try:
-                dup = self.statusmessages.index(status)
-            except ValueError:
-                pass
-            else:
-                self.statuscount[dup] += 1
-                gtk.timeout_add(timeout, lambda x: self.remove_status(status), 0)
-                return
+    def set_task_status(self, status):
+        self.task_status.pop(1)
+        self.task_status.push(1, status)
 
-        self.statusmessages.append(status)
-        message = self._get_status_message()
-        if len(self.statusmessages)==1:
-            self.appbar.push(message)
-        else:
-            self.appbar.set_status(message)
-        if timeout:
-            gtk.timeout_add(timeout, lambda x: self.remove_status(status), 0)
-        self.statuscount.append(1)
-
-    def _get_status_message(self):
-        return "[%s]" % "] [".join(self.statusmessages)
-
-    def remove_status(self, status):
-        i = self.statusmessages.index(status)
-        if self.statuscount[i] == 1:
-            self.statusmessages.pop(i)
-            self.statuscount.pop(i)
-            if len(self.statusmessages)==0:
-                self.appbar.pop()
-            else:
-                message = self._get_status_message()
-                self.appbar.set_status(message)
-        else:
-            self.statuscount[i] -= 1
+    def set_doc_status(self, status):
+        self.doc_status.pop(1)
+        self.doc_status.push(1, status)
 
 ################################################################################
 #
@@ -383,29 +393,6 @@ class NotebookLabel(gtk.HBox):
         if onclose:
             self.button.connect("clicked", onclose)
 
-
-
-################################################################################
-#
-# MeldNewMenu
-#
-################################################################################
-class MeldNewMenu(gnomeglade.Component):
-    def __init__(self, app):
-        gladefile = paths.share_dir("glade2/meldapp.glade")
-        gnomeglade.Component.__init__(self, gladefile, "popup_new")
-        self.parent = app
-    def on_menu_new_diff2_activate(self, *extra):
-        self.parent.on_menu_new_diff2_activate()
-    def on_menu_new_diff3_activate(self, *extra):
-        self.parent.on_menu_new_diff3_activate()
-    def on_menu_new_dir2_activate(self, *extra):
-        self.parent.on_menu_new_dir2_activate()
-    def on_menu_new_dir3_activate(self, *extra):
-        self.parent.on_menu_new_dir3_activate()
-    def on_menu_new_cvsview_activate(self, *extra):
-        self.parent.on_menu_new_cvsview_activate()
-
 ################################################################################
 #
 # MeldPreferences
@@ -413,9 +400,14 @@ class MeldNewMenu(gnomeglade.Component):
 ################################################################################
 class MeldPreferences(prefs.Preferences):
     defaults = {
+        "window_size_x": prefs.Value(prefs.INT, 600),
+        "window_size_y": prefs.Value(prefs.INT, 600),
         "use_custom_font": prefs.Value(prefs.BOOL,0),
         "custom_font": prefs.Value(prefs.STRING,"monospace, 14"),
         "tab_size": prefs.Value(prefs.INT, 4),
+        "show_line_numbers": prefs.Value(prefs.BOOL, 0),
+        "use_syntax_highlighting": prefs.Value(prefs.BOOL, 0),
+        "edit_wrap_lines" : prefs.Value(prefs.INT, 0),
         "edit_command_type" : prefs.Value(prefs.STRING, "internal"), #internal, gnome, custom
         "edit_command_custom" : prefs.Value(prefs.STRING, "gedit"),
         "supply_newline": prefs.Value(prefs.BOOL,1),
@@ -445,10 +437,13 @@ class MeldPreferences(prefs.Preferences):
                                               "CVS\t1\tCVS\n" + \
                                               "Binaries\t1\t*.{pyc,a,obj,o,so,la,lib,dll}\n" + \
                                               "Media\t0\t*.{jpg,gif,png,wav,mp3,ogg,xcf,xpm}"),
-        "regexes" : prefs.Value(prefs.STRING, "CVS keywords\t0\t\$[^:]+:[^\$]+\$\n" + \
-                                              "C comment\t0\t/\*[^*]*\*+([^/*][^*]*\*+)*/\n" + \
+        "regexes" : prefs.Value(prefs.STRING, "CVS keywords\t0\t\$[^:]+:[^\\n\$]+\$\n" + \
                                               "C++ comment\t0\t//.*\n" + \
-                                              "Script comment\t0\t#.*")
+                                              "C comment\t0\t/\*.*?\*/\n" + \
+                                              "All whitespace\t0\t[ \\t\\r\\f\\v]*\n" + \
+                                              "Leading whitespace\t0\t^[ \\t\\r\\f\\v]*\n" + \
+                                              "Script comment\t0\t#.*"),
+        "ignore_blank_lines" : prefs.Value(prefs.BOOL, 1)
     }
 
     def __init__(self):
@@ -478,7 +473,7 @@ class MeldPreferences(prefs.Preferences):
 
     def get_toolbar_style(self):
         if self.toolbar_style == 0:
-            style = self._gconf.get_string('/desktop/gnome/interface/toolbar_style')
+            style = self._gconf.get_string('/desktop/gnome/interface/toolbar_style') or "both"
             style = style.replace("-","_")
             style = {"both":gtk.TOOLBAR_BOTH, "text":gtk.TOOLBAR_TEXT,
                      "icon":gtk.TOOLBAR_ICONS, "icons":gtk.TOOLBAR_ICONS,
@@ -491,7 +486,7 @@ class MeldPreferences(prefs.Preferences):
 
     def get_gnome_editor_command(self, files):
         argv = []
-        editor = self._gconf.get_string('/desktop/gnome/applications/editor/exec')
+        editor = self._gconf.get_string('/desktop/gnome/applications/editor/exec') or "gedit"
         if self._gconf.get_bool("/desktop/gnome/applications/editor/needs_term"):
             texec = self._gconf.get_string("/desktop/gnome/applications/terminal/exec")
             if texec:
@@ -521,9 +516,8 @@ class MeldApp(gnomeglade.GnomeApp):
     def __init__(self):
         gladefile = paths.share_dir("glade2/meldapp.glade")
         gnomeglade.GnomeApp.__init__(self, "meld", version, gladefile, "meldapp")
-        self._map_widgets_into_lists( ["menu_file_save_file", "settings_drawstyle"] )
-        self.popup_new = MeldNewMenu(self)
-        self.statusbar = MeldStatusBar(self.appbar)
+        self._map_widgets_into_lists( "settings_drawstyle".split() )
+        self.statusbar = MeldStatusBar(self.task_progress, self.task_status, self.doc_status)
         self.prefs = MeldPreferences()
         if not developer:#hide magic testing button
             self.toolbar_magic.hide()
@@ -535,22 +529,24 @@ class MeldApp(gnomeglade.GnomeApp):
         self.idle_hooked = 0
         self.scheduler = task.LifoScheduler()
         self.scheduler.connect("runnable", self.on_scheduler_runnable )
+        self.widget.set_default_size(self.prefs.window_size_x, self.prefs.window_size_y)
+        self.widget.show()
 
     def on_idle(self):
         ret = self.scheduler.iteration()
         if ret:
-            if type(ret) == type(""):
-                self.appbar.set_status(ret)
+            if type(ret) in (type(""), type(u"")):
+                self.statusbar.set_task_status(ret)
             elif type(ret) == type(0.0):
-                self.appbar.get_progress().set_fraction(ret)
+                self.statusbar.task_progress.set_fraction(ret)
             else:
-                self.appbar.get_progress().pulse()
+                self.statusbar.task_progress.pulse()
         else:
-                self.appbar.pop()
-                self.appbar.get_progress().set_fraction(0)
+            self.statusbar.task_progress.set_fraction(0)
         if self.scheduler.tasks_pending():
             return 1
         else:
+            self.statusbar.set_task_status("")
             self.idle_hooked = 0
             return 0
 
@@ -574,11 +570,10 @@ class MeldApp(gnomeglade.GnomeApp):
         newseq = newdoc.undosequence
         self.button_undo.set_sensitive(newseq.can_undo())
         self.button_redo.set_sensitive(newseq.can_redo())
-        for i in range(3):
-            sensitive = newdoc.num_panes > i
-            self.menu_file_save_file[i].set_sensitive(sensitive)
         nbl = self.notebook.get_tab_label( newdoc.widget )
         self.widget.set_title( nbl.label.get_text() + " - Meld")
+        self.statusbar.set_doc_status("")
+        newdoc.on_switch_event()
         self.scheduler.add_task( newdoc.scheduler )
 
     def on_notebook_label_changed(self, component, text):
@@ -592,37 +587,19 @@ class MeldApp(gnomeglade.GnomeApp):
 
     def on_can_redo(self, undosequence, can):
         self.button_redo.set_sensitive(can)
-    
+
+    def on_size_allocate(self, window, rect):
+        self.prefs.window_size_x = rect.width
+        self.prefs.window_size_y = rect.height
+
     #
     # Toolbar and menu items (file)
     #
-    def on_menu_new_diff2_activate(self, *extra):
-        BrowseFileDialog(self, [_("Original File"), _("Modified File")], self.append_filediff)
-
-    def on_menu_new_diff3_activate(self, *extra):
-        BrowseFileDialog(self, [_("Other Changes"), _("Common Ancestor"), _("Local Changes")], self.append_filediff )
-
-    def on_menu_new_dir2_activate(self, *extra):
-        BrowseFileDialog(self, [_("Original Directory"), _("Modified Directory")], self.append_dirdiff, isdir=1)
-
-    def on_menu_new_dir3_activate(self, *extra):
-        BrowseFileDialog(self, [_("Other Directory"), _("Original Directory"), _("Modified Directory")], self.append_dirdiff, isdir=1)
-
-    def on_menu_new_cvsview_activate(self, *extra):
-        BrowseFileDialog(self,[_("Root CVS Directory")], self.append_cvsview, isdir=1)
+    def on_menu_file_new_activate(self, menuitem):
+        NewDocDialog(self, NewDocDialog.TYPE.DIFF2)
 
     def on_menu_save_activate(self, menuitem):
-        try:
-            index = self.menu_file_save_file.index(menuitem)
-        except ValueError:
-            index = -1
-        try:
-            if index >= 0: # save one
-                self.current_doc().save_file(index)
-            else: # save all
-                self.current_doc().save()
-        except AttributeError:
-            pass
+        self.current_doc().save()
 
     def on_menu_refresh_activate(self, *args):
         self.current_doc().refresh()
@@ -708,14 +685,14 @@ class MeldApp(gnomeglade.GnomeApp):
             #self.append_filediff( ("ntest/file9a", "ntest/file9b") )
         pass
 
-    def on_menu_down_activate(self, *args):
+    def on_menu_edit_down_activate(self, *args):
         misc.safe_apply( self.current_doc(), "next_diff", gtk.gdk.SCROLL_DOWN )
 
-    def on_menu_up_activate(self, *args):
+    def on_menu_edit_up_activate(self, *args):
         misc.safe_apply( self.current_doc(), "next_diff", gtk.gdk.SCROLL_UP )
 
     def on_toolbar_new_clicked(self, *args):
-        self.popup_new.widget.popup(None, None, None, 3, gtk.get_current_event_time() )
+        NewDocDialog(self, NewDocDialog.TYPE.DIFF2)
 
     def on_toolbar_stop_clicked(self, *args):
         self.current_doc().stop()
@@ -742,6 +719,7 @@ class MeldApp(gnomeglade.GnomeApp):
         page.connect("label-changed", self.on_notebook_label_changed)
         page.connect("file-changed", self.on_file_changed)
         page.connect("create-diff", lambda obj,arg: self.append_filediff(arg) )
+        page.connect("status-changed", lambda junk,arg: self.statusbar.set_doc_status(arg) )
 
     def append_dirdiff(self, dirs):
         assert len(dirs) in (1,2,3)
@@ -759,6 +737,24 @@ class MeldApp(gnomeglade.GnomeApp):
         self._append_page(doc, "tree-file-normal.png")
         doc.set_files(files)
 
+    def append_diff(self, paths):
+        aredirs = [ os.path.isdir(p) for p in paths ]
+        arefiles = [ os.path.isfile(p) for p in paths ]
+        if (1 in aredirs) and (1 in arefiles):
+            m = _("Cannot compare a mixture of files and directories.\n")
+            for i in range(len(paths)):
+                status = aredirs[i] and _("folder") \
+                    or arefiles[i] and _("file") \
+                    or _("nonexistant")
+                m += "\n(%s)\t`%s'" % (status, paths[i])
+            misc.run_dialog( m,
+                    parent = self,
+                    buttonstype = gtk.BUTTONS_OK)
+        elif 1 in aredirs:
+            self.append_dirdiff(paths)
+        else:
+            self.append_filediff(paths)
+
     def append_cvsview(self, locations):
         assert len(locations) in (1,)
         location = locations[0]
@@ -766,6 +762,13 @@ class MeldApp(gnomeglade.GnomeApp):
         self._append_page(doc, "cvs-icon.png")
         doc.set_location(location)
 
+    def append_svnview(self, locations):
+        assert len(locations) in (1,)
+        location = locations[0]
+        doc = svnview.CvsView(self.prefs)
+        self._append_page(doc, "svn-icon.png")
+        doc.set_location(location)
+
     #
     # Current doc actions
     #
@@ -805,12 +808,17 @@ usage_string = _("""Meld is a file and directory comparison tool. Usage:
     meld <file> <file> [file]   Start with 2 or 3 way file comparison
     meld <dir>  <dir>  [dir]    Start with 2 or 3 way directory comparison
 
+Options:
+    -h, --help                  Show this help text and exit
+    -v, --version               Display the version and exit
+
 For more information choose help -> contents.
-Report bugs to steve9000@users.sourceforge.net.
+Report bugs at http://bugzilla.gnome.org/buglist.cgi?product=meld
+Discuss meld at http://mail.gnome.org/mailman/listinfo/gnome-devtools
 """)
 
 version_string = _("""Meld %s
-Written by Stephen Kennedy <steve9000@users.sf.net>""") % version
+Written by Stephen Kennedy <stevek@gnome.org>""") % version
 
 ################################################################################
 #
@@ -827,64 +835,42 @@ def main():
         def __getattr__(self, attr):
             return getattr(self.file, attr)
     sys.stdout = Unbuffered(sys.stdout)
+    args = sys.argv[1:]
 
-    if len(sys.argv) == 2:
-        if sys.argv[1] in ("-h", "--help"):
+    if len(args) >= 2:
+        if "-h" in args or "--help" in args:
             print usage_string
             return
-        elif sys.argv[1] in ("-v", "--version"):
+        if "-v" in args or "--version" in args:
             print version_string
             return
 
     app = MeldApp()
-    arg = sys.argv[1:]
 
-    if len(arg) == 0:
+    if len(args) == 0:
         pass
 
-    elif len(arg) == 1:
-        a = arg[0]
-        if os.path.isdir(a):
-            app.append_cvsview( [a] )
-        elif os.path.isfile(a):
+    elif len(args) == 1:
+        a = args[0]
+        if os.path.isfile(a):
             doc = cvsview.CvsView(app.prefs)
             def cleanup():
                 app.scheduler.remove_scheduler(doc.scheduler)
             app.scheduler.add_task(cleanup)
             app.scheduler.add_scheduler(doc.scheduler)
             doc.set_location( os.path.dirname(a) )
-            doc.connect("create-diff", lambda obj,arg: app.append_filediff(arg) )
+            doc.connect("create-diff", lambda obj,arg: app.append_diff(arg) )
             doc.run_cvs_diff([a])
         else:
-            app.usage( _("`%s' is not a directory or file, cannot open cvs view") % arg[0])
+            if os.path.exists( os.path.join(a,".svn") ):
+                app.append_svnview( [a] )
+            else:
+                app.append_cvsview( [a] )
                 
-    elif len(arg) in (2,3):
-        done = 0
-        exists = map( lambda a: os.access(a, os.R_OK), arg)
-        if 0 in exists:
-            m = _("Cannot open ")
-            for i in range(len(arg)):
-                if not exists[i]:
-                    m += "`%s'" % arg[i]
-            app.usage(m)
-            done = 1
-        if not done:
-            arefiles = map( os.path.isfile, arg)
-            if 0 not in arefiles:
-                app.append_filediff( arg )
-                done = 1
-        if not done:
-            aredirs = map( os.path.isdir, arg)
-            if 0 not in aredirs:
-                app.append_dirdiff( arg )
-                done = 1
-        if not done:
-            m = _("Cannot compare a mixture of files and directories.\n")
-            for i in range(len(arg)):
-                m += "(%s)\t`%s'\n" % (arefiles[i] and _("file") or _("folder"), arg[i])
-            app.usage(m)
+    elif len(args) in (2,3):
+        app.append_diff(args)
     else:
-        app.usage( _("Wrong number of arguments (Got %i)") % len(arg))
+        app.usage( _("Wrong number of arguments (Got %i)") % len(args))
 
-    app.mainloop()
+    app.main()
 
