@@ -1,4 +1,4 @@
-### Copyright (C) 2002-2003 Stephen Kennedy <stevek@gnome.org>
+### Copyright (C) 2002-2004 Stephen Kennedy <stevek@gnome.org>
 
 ### This program is free software; you can redistribute it and/or modify
 ### it under the terms of the GNU General Public License as published by
@@ -20,6 +20,7 @@ import calendar
 import tempfile
 import gobject
 import shutil
+import errno
 import time
 import copy
 import gtk
@@ -80,22 +81,29 @@ def _lookup_cvs_files(dirs, files):
     else:
         return [],[]
 
-    entries = os.popen("svn status -Nv "+directory).read()
+    while 1:
+        try:
+            entries = os.popen("svn status -Nv "+directory).read()
+            break
+        except OSError, e:
+            if e.errno != errno.EAGAIN:
+                raise
 
     retfiles = []
     retdirs = []
-    matches = re.findall("^(.)....\s*\d*\s*(\d*) [^ ]*\s*(.*)$(?m)", entries)
+    matches = re.findall("^(.)....\s*\d*\s*(\d*)\s*\w*\s*(.*?)$(?m)", entries)
+    matches = [ (m[2].split()[-1],m[0],m[1]) for m in matches]
     matches.sort()
 
     for match in matches:
-        name = match[2]
-        if(match[0] == "!" or match[0] == "A"):
+        name = match[0]
+        if(match[1] == "!" or match[1] == "A"):
             # for new or missing files, the findall expression
             # does not supply the correct name 
             name = re.sub(r'^[?]\s*(.*)$', r'\1', name)
         isdir = os.path.isdir(name)
         path = os.path.join(directory, name)
-        rev = match[1]
+        rev = match[2]
         options = ""
         tag = ""
         if tag:
@@ -109,22 +117,13 @@ def _lookup_cvs_files(dirs, files):
             if(os.path.basename(name) != os.path.basename(directory)):
                 retdirs.append( Dir(path,name,state) )
         else:
-            state=0
-            if(match[0] == "?"):
-                state = tree.STATE_NONE
-            if(match[0] == "A"):
-                state = tree.STATE_NEW
-            elif(match[0] == " "):
-                state = tree.STATE_NORMAL                
-            elif(match[0] == "!"):
-                state = tree.STATE_MISSING
-            elif(match[0] == "I"):
-                state = tree.STATE_IGNORED                
-            elif(match[0] == "M"):
-                state = tree.STATE_MODIFIED
-            elif(match[0] == "C"):
-                state = tree.STATE_CONFLICT
-
+            state = { "?": tree.STATE_NONE,
+                      "A": tree.STATE_NEW,
+                      " ": tree.STATE_NORMAL,
+                      "!": tree.STATE_MISSING,
+                      "I": tree.STATE_IGNORED,
+                      "M": tree.STATE_MODIFIED,
+                      "C": tree.STATE_CONFLICT }.get(match[1], tree.STATE_NONE)
             retfiles.append( File(path, name, state, rev, tag, options) )
 
     return retdirs, retfiles
@@ -397,6 +396,10 @@ class CvsView(melddoc.MeldDoc, gnomeglade.Component):
             else: # just the root
                 self.treeview.expand_row( (0,), 0)
 
+    def on_preference_changed(self, key, value):
+        if key == "toolbar_style":
+            self.toolbar.set_style( self.prefs.get_toolbar_style() )
+
     def on_fileentry_activate(self, fileentry):
         path = fileentry.get_full_path(0)
         self.set_location(path)
@@ -454,14 +457,22 @@ class CvsView(melddoc.MeldDoc, gnomeglade.Component):
     def on_button_filter_toggled(self, button):
         self.refresh()
 
-    def _get_selected_files(self):
-        ret = []
+    def _get_selected_treepaths(self):
+        sel = []
         def gather(model, path, iter):
-            ret.append( model.value_path(iter,0) )
+            sel.append( model.get_path(iter) )
+        s = self.treeview.get_selection()
+        s.selected_foreach(gather)
+        return sel
+
+    def _get_selected_files(self):
+        sel = []
+        def gather(model, path, iter):
+            sel.append( model.value_path(iter,0) )
         s = self.treeview.get_selection()
         s.selected_foreach(gather)
         # remove empty entries and remove trailing slashes
-        return map(lambda x: x[-1]!="/" and x or x[:-1], filter(lambda x: x!=None, ret))
+        return [ x[-1]!="/" and x or x[:-1] for x in sel if x != None ]
 
     def _command_iter(self, command, files, refresh):
         """Run 'command' on 'files'. Return a tuple of the directory the
@@ -470,14 +481,14 @@ class CvsView(melddoc.MeldDoc, gnomeglade.Component):
         msg = misc.shelljoin(command)
         yield "[%s] %s" % (self.label_text, msg.replace("\n", u"\u21b2") )
         if len(files) == 1 and os.path.isdir(files[0]):
-            workdir = files[0]
-            files = ["."]
+            workdir = os.path.dirname( files[0] )
+            files = [ os.path.basename( files[0] ) ]
         else:
             workdir = _commonprefix(files)
-        kill = len(workdir) and (len(workdir)+1) or 0
-        files = filter(lambda x: len(x), map(lambda x: x[kill:], files))
+            kill = len(workdir) and (len(workdir)+1) or 0
+            files = filter(lambda x: len(x), map(lambda x: x[kill:], files))
         r = None
-        self.consolestream.write( misc.shelljoin(command+files) + "(in %s)\n" % workdir)
+        self.consolestream.write( misc.shelljoin(command+files) + " (in %s)\n" % workdir)
         readfunc = misc.read_pipe_iter(command + files, self.consolestream, workdir=workdir).next
         try:
             while r == None:
@@ -677,5 +688,21 @@ class CvsView(melddoc.MeldDoc, gnomeglade.Component):
         item.show()
         menu.insert( item, 1 )
 
+    def next_diff(self, direction):
+        start_iter = self.model.get_iter( (self._get_selected_treepaths() or [(0,)])[-1] )
+
+        def goto_iter(it):
+            curpath = self.model.get_path(it)
+            for i in range(len(curpath)-1):
+                self.treeview.expand_row( curpath[:i+1], 0)
+            self.treeview.set_cursor(curpath)
+
+        search = {gtk.gdk.SCROLL_UP : self.model.inorder_search_up}.get(direction, self.model.inorder_search_down)
+        for it in search( start_iter ):
+            state = int(self.model.get_state( it, 0))
+            if state not in (tree.STATE_NORMAL, tree.STATE_EMPTY):
+                goto_iter(it)
+                return
+
 gobject.type_register(CvsView)
 
